"""Post processing API."""

import atooms.postprocessing as postprocessing
from atooms.postprocessing.partial import Partial
from atooms.trajectory import Trajectory
from atooms.trajectory.decorators import change_species
from atooms.system.particle import distinct_species

from .helpers import linear_grid, logx_grid


def _get_trajectories(input_files, args):
    from atooms.trajectory import Sliced
    from atooms.core.utils import fractional_slice
    for input_file in input_files:
        with Trajectory(input_file, fmt=args['fmt']) as th:
            # Caching is useful for systems with multiple species but
            # it will increase the memory footprint. Use --no-cache to
            # disable it
            if not args['no_cache']:
                th.cache = True
            if args['species_layout'] is not None:
                th.register_callback(change_species, args['species_layout'])
            sl = fractional_slice(args['first'], args['last'], args['skip'], len(th))
            if th.block_size > 1:
                sl_start = (sl.start // th.block_size) * th.block_size if sl.start is not None else sl.start
                sl_stop = (sl.stop // th.block_size) * th.block_size if sl.stop is not None else sl.stop
                sl = slice(sl_start, sl_stop, sl.step)
            ts = Sliced(th, sl)
            yield ts

def _compat(args, fmt, species_layout=None):
    if 'first' not in args:
        args['first'] = None
    if 'last' not in args:
        args['last'] = None
    if 'skip' not in args:
        args['skip'] = None
    if 'fmt' not in args:
        args['fmt'] = fmt
    if 'species_layout' not in args:
        args['species_layout'] = species_layout
    if 'norigins' not in args:
        args['norigins'] = None
    if 'fast' not in args:
        args['fast'] = False
    if 'no_cache' not in args:
        args['no_cache'] = False
    return args

def gr(input_file, dr=0.04, grandcanonical=False, fmt=None, species_layout=None,
       *input_files, **global_args):
    """Radial distribution function"""
    global_args = _compat(global_args, fmt=fmt, species_layout=species_layout)
    for th in _get_trajectories([input_file] + list(input_files), global_args):
        th._grandcanonical = grandcanonical
        postprocessing.RadialDistributionFunction(th, dr=dr, norigins=global_args['norigins']).do()
        ids = distinct_species(th[-1].particle)
        if len(ids) > 1:
            Partial(postprocessing.RadialDistributionFunction, ids, th, dr=dr, norigins=global_args['norigins']).do()

def sk(input_file, nk=20, dk=0.1, kmin=-1.0, kmax=15.0, ksamples=30,
       species_layout=None, fmt=None, trajectory_field=None,
       field=None, *input_files, **global_args):
    """Structure factor"""
    global_args = _compat(global_args, fmt=fmt, species_layout=species_layout)
    if global_args['fast']:
        backend = postprocessing.StructureFactorOpti
    else:
        backend = postprocessing.StructureFactor

    for th in _get_trajectories([input_file] + list(input_files), global_args):
        ids = distinct_species(th[-1].particle)
        backend(th, None, norigins=global_args['norigins'],
                trajectory_field=trajectory_field,
                field=field, kmin=kmin,
                kmax=kmax, nk=nk, dk=dk,
                ksamples=ksamples).do()
        if len(ids) > 1 and trajectory_field is None:
            Partial(backend, ids, th, None,
                    norigins=global_args['norigins'], kmin=kmin,
                    kmax=kmax, nk=nk, dk=dk,
                    ksamples=ksamples).do()

def ik(input_file, trajectory_radius=None, nk=20, dk=0.1, kmin=-1.0, kmax=15.0,
       ksamples=30, fmt=None, species_layout=None,
       *input_files, **global_args):
    """Spectral density"""
    global_args = _compat(global_args, fmt=fmt, species_layout=species_layout)
    for th in _get_trajectories([input_file] + list(input_files), global_args):
        if trajectory_radius is None:
            trajectory_radius = input_file
            postprocessing.SpectralDensity(th, trajectory_radius,
                                           kgrid=None, norigins=global_args['norigins'],
                                           kmin=kmin, kmax=kmax, nk=nk, dk=dk,
                                           ksamples=ksamples).do()

def msd(input_file, time_target=-1.0, time_target_fraction=0.75,
        tsamples=30, sigma=1.0, func=linear_grid, rmsd_target=-1.0,
        fmt=None, species_layout=None, *input_files, **global_args):
    """Mean square displacement"""
    global_args = _compat(global_args, fmt=fmt, species_layout=species_layout)
    for th in _get_trajectories([input_file] + list(input_files), global_args):
        dt = th.timestep
        if time_target > 0:
            t_grid = [0.0] + func(dt, min(th.total_time, time_target), tsamples)
        elif time_target_fraction > 0:
            t_grid = [0.0] + func(dt, time_target_fraction*th.total_time, tsamples)
        else:
            t_grid = None
        ids = distinct_species(th[-1].particle)
        postprocessing.MeanSquareDisplacement(th, tgrid=t_grid,
                                              norigins=global_args['norigins'],
                                              sigma=sigma, rmax=rmsd_target).do()
        if len(ids) > 1:
            Partial(postprocessing.MeanSquareDisplacement, ids,
                    th, tgrid=t_grid, norigins=global_args['norigins'], sigma=sigma, rmax=rmsd_target).do()

def vacf(input_file, time_target=-1.0, time_target_fraction=0.10,
         tsamples=30, func=linear_grid, fmt=None, species_layout=None,
         *input_files, **global_args):
    """Velocity autocorrelation function"""
    global_args = _compat(global_args, fmt=fmt, species_layout=species_layout)
    for th in _get_trajectories([input_file] + list(input_files), global_args):
        if time_target > 0:
            t_grid = [0.0] + func(th.timestep, min(th.total_time, time_target), tsamples)
        elif time_target_fraction > 0:
            t_grid = [0.0] + func(th.timestep, time_target_fraction*th.total_time, tsamples)
        else:
            t_grid = None
        postprocessing.VelocityAutocorrelation(th, t_grid,
                                               norigins=global_args['norigins']).do()
        ids = distinct_species(th[-1].particle)
        if len(ids) > 1:
            Partial(postprocessing.VelocityAutocorrelation, ids, th,
                    t_grid, norigins=global_args['norigins']).do()

def fkt(input_file, time_target=-1.0, time_target_fraction=0.75,
        tsamples=60, kmin=7.0, kmax=7.0, ksamples=1, dk=0.1, nk=100,
        tag_by_name=False, func=logx_grid, fmt=None,
        species_layout=None, *input_files, **global_args):
    """Total intermediate scattering function"""
    global_args = _compat(global_args, fmt=fmt, species_layout=species_layout)
    for th in _get_trajectories([input_file] + list(input_files), global_args):
        if time_target > 0:
            t_grid = [0.0] + func(th.timestep, time_target, tsamples)
        elif time_target_fraction > 0:
            t_grid = [0.0] + func(th.timestep, time_target_fraction*th.total_time, tsamples)
        else:
            t_grid = None
        k_grid = linear_grid(kmin, kmax, ksamples)
        ids = distinct_species(th[0].particle)
        if len(ids) > 1:
            Partial(postprocessing.IntermediateScattering, ids, th, k_grid, t_grid,
                    nk=nk, dk=dk).do()

def fskt(input_file, time_target=-1.0, time_target_fraction=0.75,
         tsamples=60, kmin=7.0, kmax=8.0, ksamples=1, dk=0.1, nk=8,
         func=logx_grid, fmt=None, species_layout=None, total=False,
         *input_files, **global_args):
    """Self intermediate scattering function"""
    global_args = _compat(global_args, fmt=fmt, species_layout=species_layout)
    for th in _get_trajectories([input_file] + list(input_files), global_args):
        if time_target > 0:
            t_grid = [0.0] + func(th.timestep, time_target, tsamples)
        elif time_target_fraction > 0:
            t_grid = [0.0] + func(th.timestep, time_target_fraction*th.total_time, tsamples)
        else:
            t_grid = None
        k_grid = linear_grid(kmin, kmax, ksamples)
        if total:
            postprocessing.SelfIntermediateScattering(th, k_grid, t_grid,
                                                      nk, dk=dk, norigins=global_args['norigins']).do()
        ids = distinct_species(th[-1].particle)
        if len(ids) > 1:
            Partial(postprocessing.SelfIntermediateScattering, ids,
                    th, k_grid, t_grid, nk, dk=dk, norigins=global_args['norigins']).do()

def chi4qs(input_file, tsamples=60, a=0.3, time_target=-1.0,
           time_target_fraction=0.75, fmt=None, species_layout=None,
           total=False, *input_files, **global_args):
    """Dynamic susceptibility of self overlap"""
    global_args = _compat(global_args, fmt=fmt, species_layout=species_layout)

    if global_args['fast']:
        backend = postprocessing.Chi4SelfOverlapOpti
    else:
        backend = postprocessing.Chi4SelfOverlap

    for th in _get_trajectories([input_file] + list(input_files), global_args):
        func = logx_grid
        if time_target > 0:
            t_grid = [0.0] + func(th.timestep, min(th.total_time, time_target), tsamples)
        elif time_target_fraction > 0:
            t_grid = [0.0] + func(th.timestep, time_target_fraction*th.total_time, tsamples)
        else:
            t_grid = None
        if total:
            backend(th, t_grid, a=a, norigins=global_args['norigins']).do()
        ids = distinct_species(th[0].particle)
        if not total and len(ids) > 1:
            Partial(backend, ids, th, t_grid, a=a, norigins=global_args['norigins']).do()

def alpha2(input_file, time_target=-1.0, time_target_fraction=0.75,
           tsamples=60, func=logx_grid, fmt=None,
         species_layout=None, *input_files, **global_args):
    """Non-Gaussian parameter"""
    global_args = _compat(global_args, fmt=fmt, species_layout=species_layout)
    for th in _get_trajectories([input_file] + list(input_files), global_args):
        if time_target > 0:
            t_grid = [0.0] + func(th.timestep, time_target, tsamples)
        elif time_target_fraction > 0:
            t_grid = [0.0] + func(th.timestep, time_target_fraction*th.total_time, tsamples)
        else:
            t_grid = None
        postprocessing.NonGaussianParameter(th, t_grid, norigins=global_args['norigins']).do()
        ids = distinct_species(th[-1].particle)
        if len(ids) > 1:
            Partial(postprocessing.NonGaussianParameter, ids, th, t_grid, norigins=global_args['norigins']).do()



# This file is part of atooms
# Copyright 2010-2018, Daniele Coslovich

"""Structure factor."""

import logging
import numpy

from .progress import progress
from .fourierspace import FourierSpaceCorrelation, expo_sphere

__all__ = ['StructureFactor', 'StructureFactorOptimized']

_log = logging.getLogger(__name__)


class StructureFactor(FourierSpaceCorrelation):
    """
    Structure factor.

    If `trajectory_field` is not `None`, the field is read from the
    last column of this trajectory file, unless the `field` string is
    provided.

    See the documentation of the `FourierSpaceCorrelation` base class
    for information on the instance variables.
    """

    nbodies = 2
    symbol = 'sk'
    short_name = 'S(k)'
    long_name = 'structure factor'
    phasespace = ['pos']

    def __init__(self, trajectory, kgrid=None, norigins=-1, nk=20,
                 dk=0.1, kmin=-1.0, kmax=15.0, ksamples=30,
                 trajectory_field=None, field=None):
        FourierSpaceCorrelation.__init__(self, trajectory, kgrid, norigins,
                                         nk, dk, kmin, kmax, ksamples)
        self._is_cell_variable = None
        self._field, tag = self._add_field(trajectory_field, field)
        if tag is not None:
            self.tag = tag
            self.tag_description += '%s field' % tag.replace('_', ' ')

    def _add_field(self, trajectory_field, field):
        if trajectory_field is None:
            return None, None
        else:
            # TODO: check step consistency 06.09.2017
            from atooms.trajectory import TrajectoryXYZ
            with TrajectoryXYZ(trajectory_field) as th:
                if th.steps != self.trajectory.steps:
                    raise ValueError('field and traectory are not synced (%s, %s)' % (len(th), len(self.trajectory)))
                fields = []
                # This must be a string, not a list
                unique_field = th._read_metadata(0)['columns']
                if isinstance(unique_field, list):
                    # If field is not given, get the last column
                    if field is None:
                        unique_field = unique_field[-1]
                    else:
                        unique_field = field
                for s in th:
                    fields.append(s.dump('particle.%s' % unique_field))
            return fields, unique_field

    def _compute(self):
        from atooms.trajectory.utils import is_cell_variable

        nsteps = len(self._pos_0)
        # Setup k vectors and tabulate rho
        kgrid, selection = self.kgrid, self.selection
        kmax = max(self.kvector.keys()) + self.dk
        cnt = [0 for k in kgrid]
        rho_av = [complex(0., 0.) for k in kgrid]
        rho2_av = [complex(0., 0.) for k in kgrid]
        variable_cell = is_cell_variable(self.trajectory)
        for i in progress(range(0, nsteps, self.skip), total=nsteps // self.skip):
            # If cell changes we have to update the wave vectors
            if variable_cell:
                self._setup(i)
                kgrid, selection = self._decimate_k()
                kmax = max(self.kvector.keys()) + self.dk

            # Tabulate exponentials
            # Note: tabulating and computing takes about the same time
            if self._pos_0[i] is self._pos_1[i]:
                # Identical species
                expo_0 = expo_sphere(self.k0, kmax, self._pos_0[i])
                expo_1 = expo_0
            else:
                # Cross correlation
                expo_0 = expo_sphere(self.k0, kmax, self._pos_0[i])
                expo_1 = expo_sphere(self.k0, kmax, self._pos_1[i])

            for kk, knorm in enumerate(kgrid):
                for k in selection[kk]:
                    ik = self.kvector[knorm][k]
                    # In the absence of a microscopic field, rho_av = (0, 0)
                    if not self._field:
                        if expo_0 is expo_1:
                            # Identical species
                            rho_0 = numpy.sum(expo_0[..., 0, ik[0]] *
                                              expo_0[..., 1, ik[1]] *
                                              expo_0[..., 2, ik[2]])
                            rho_1 = rho_0
                        else:
                            # Cross correlation
                            rho_0 = numpy.sum(expo_0[..., 0, ik[0]] *
                                              expo_0[..., 1, ik[1]] *
                                              expo_0[..., 2, ik[2]])
                            rho_1 = numpy.sum(expo_1[..., 0, ik[0]] *
                                              expo_1[..., 1, ik[1]] *
                                              expo_1[..., 2, ik[2]])
                    else:
                        # We have a field as a weight
                        rho_0 = numpy.sum(self._field[i] *
                                          expo_0[..., 0, ik[0]] *
                                          expo_0[..., 1, ik[1]] *
                                          expo_0[..., 2, ik[2]])
                        rho_1 = rho_0
                        rho_av[kk] += rho_0

                    rho2_av[kk] += (rho_0 * rho_1.conjugate())
                    cnt[kk] += 1

        # Normalization.
        npart_0 = sum([p.shape[0] for p in self._pos_0]) / float(len(self._pos_0))
        npart_1 = sum([p.shape[0] for p in self._pos_1]) / float(len(self._pos_1))
        self.grid = kgrid
        self.value, self.value_nonorm = [], []
        for kk in range(len(self.grid)):
            norm = float(npart_0 * npart_1)**0.5
            value = (rho2_av[kk] / cnt[kk] - rho_av[kk]*rho_av[kk].conjugate() / cnt[kk]**2).real
            self.value.append(value / norm)
            self.value_nonorm.append(value)


class StructureFactorOptimized(StructureFactor):
    """
    Optimized structure factor.

    It uses a fortran 90 extension.
    """

    nbodies = 2
    symbol = 'sk'
    short_name = 'S(k)'
    long_name = 'structure factor'
    phasespace = ['pos']

    def _compute(self):
        from atooms.trajectory.utils import is_cell_variable
        try:
            from atooms.postprocessing.fourierspace_wrap import fourierspace_module
        except ImportError:
            _log.error('f90 wrapper missing or not functioning')
            raise

        nsteps = len(self._pos_0)
        # Setup k vectors and tabulate rho
        kgrid, selection = self.kgrid, self.selection
        kmax = max(self.kvector.keys()) + self.dk
        cnt = [0 for k in kgrid]
        rho_av = [complex(0., 0.) for k in kgrid]
        rho2_av = [complex(0., 0.) for k in kgrid]
        variable_cell = is_cell_variable(self.trajectory)
        for i in range(0, nsteps, self.skip):
            # If cell changes we have to update the wave vectors
            if variable_cell:
                self._setup(i)
                kgrid, selection = self._decimate_k()
                kmax = max(self.kvector.keys()) + self.dk

            # Tabulate exponentials
            # Note: tabulating and computing takes about the same time
            if self._pos_0[i] is self._pos_1[i]:
                # Identical species
                expo_0 = expo_sphere(self.k0, kmax, self._pos_0[i])
                expo_1 = expo_0
            else:
                # Cross correlation
                # TODO: cross correlation wont work
                expo_0 = expo_sphere(self.k0, kmax, self._pos_0[i])
                expo_1 = expo_sphere(self.k0, kmax, self._pos_1[i])

            for kk, knorm in enumerate(kgrid):
                ikvec = numpy.ndarray((3, len(selection[kk])), order='F', dtype=numpy.int32)
                i = 0
                for k in selection[kk]:
                    ikvec[:, i] = self.kvector[knorm][k]
                    i += 1
                rho = numpy.zeros(ikvec.shape[1], dtype=numpy.complex128)
                fourierspace_module.sk_bare(expo_0, ikvec, rho)
                rho_0 = rho
                rho_1 = rho
                rho2_av[kk] += numpy.sum(rho_0 * rho_1.conjugate())
                cnt[kk] += rho.shape[0]

        # Normalization.
        npart_0 = sum([p.shape[0] for p in self._pos_0]) / float(len(self._pos_0))
        npart_1 = sum([p.shape[0] for p in self._pos_1]) / float(len(self._pos_1))
        self.grid = kgrid
        self.value, self.value_nonorm = [], []
        for kk in range(len(self.grid)):
            norm = float(npart_0 * npart_1)**0.5
            value = (rho2_av[kk] / cnt[kk] - rho_av[kk]*rho_av[kk].conjugate() / cnt[kk]**2).real
            self.value.append(value / norm)
            self.value_nonorm.append(value)





# This file is part of atooms
# Copyright 2010-2018, Daniele Coslovich

"""Radial distribution function."""

import math

import numpy

from .helpers import linear_grid
from .correlation import Correlation
from .progress import progress

__all__ = ['RadialDistributionFunction']


def gr_kernel(x, y, L):
    # r is an array of array distances
    r = x-y
    r = r - numpy.rint(r/L) * L
    return numpy.sqrt(numpy.sum(r**2, axis=1))


def gr_kernel_square(x, y, L):
    """Return square distances."""
    # r is an array of array distances
    r = x-y
    r = r - numpy.rint(r/L) * L
    return numpy.sum(r**2, axis=1)


def pairs_newton_hist(f, x, y, L, bins):
    """
    Apply function f to all pairs in x[i] and y[j] and update the
    |hist| histogram using the |bins| bin edges.
    """
    hist, bins = numpy.histogram([], bins)
    # Do the calculation in batches to optimize
    bl = max(1, int(100 * 1000.0 / len(y)))
    for ib in range(0, len(y)-1, bl):
        fxy = []
        # batch must never exceed len(y)-1
        for i in range(ib, min(ib+bl, len(y)-1)):
            for value in f(x[i+1:], y[i], L):
                fxy.append(value)
        hist_tmp, bins = numpy.histogram(fxy, bins)
        hist += hist_tmp
    return hist


def pairs_hist(f, x, y, L, bins):
    """
    Apply function f to all pairs in x[i] and y[j] and update the
    |hist| histogram using the |bins| bin edges.
    """
    hist, bins = numpy.histogram([], bins)
    for i in range(len(y)):
        fxy = f(x[:], y[i], L)
        hist_tmp, bins = numpy.histogram(fxy, bins)
        hist += hist_tmp
    return hist


class RadialDistributionFunction(Correlation):
    """
    Radial distribution function.

    The correlation function g(r) is computed over a grid of distances
    `rgrid`. If the latter is `None`, the grid is linear from 0 to L/2
    with a spacing of `dr`. Here, L is the side of the simulation cell
    along the x axis at the first step.

    Additional parameters:
    ----------------------

    - norigins: controls the number of trajectory frames to compute
      the time average
    """

    nbodies = 2
    symbol = 'gr'
    short_name = 'g(r)'
    long_name = 'radial distribution function'
    phasespace = 'pos'

    def __init__(self, trajectory, rgrid=None, norigins=None, dr=0.04):
        Correlation.__init__(self, trajectory, rgrid, norigins=norigins)
        self.side = self.trajectory.read(0).cell.side
        if rgrid is not None:
            # Reconstruct bounds of grid for numpy histogram
            self.grid = []
            for i in range(len(rgrid)):
                self.grid.append(rgrid[i] - (rgrid[1] - rgrid[0]) / 2)
            self.grid.append(rgrid[-1] + (rgrid[1] - rgrid[0]) / 2)
        else:
            if len(self.side.shape) > 0:
                L = self.side[0]
            else:
                L = self.side
            self.grid = linear_grid(0.0, L / 2.0, dr)

    def _compute(self):
        ncfg = len(self.trajectory)
        if self.trajectory.grandcanonical:
            N_0 = numpy.average([len(x) for x in self._pos_0])
            N_1 = numpy.average([len(x) for x in self._pos_1])
        else:
            N_0, N_1 = len(self._pos_0[0]), len(self._pos_1[0])

        gr_all = []
        _, r = numpy.histogram([], bins=self.grid)
        origins = range(0, ncfg, self.skip)
        for i in progress(origins):
            self.side = self.trajectory.read(i).cell.side
            if len(self._pos_0[i]) == 0 or len(self._pos_1[i]) == 0:
                continue
            if self._pos_0 is self._pos_1:
                gr = pairs_newton_hist(gr_kernel, self._pos_0[i], self._pos_1[i],
                                       self.side, r)
            else:
                gr = pairs_hist(gr_kernel, self._pos_0[i], self._pos_1[i],
                                self.side, r)
            gr_all.append(gr)

        # Normalization
        vol = 4 * math.pi / 3.0 * (r[1:]**3-r[:-1]**3)
        rho = N_1 / self.side.prod()
        if self._pos_0 is self._pos_1:
            norm = rho * vol * N_0 * 0.5  # use Newton III
        else:
            norm = rho * vol * N_0
        gr = numpy.average(gr_all, axis=0)
        self.grid = (r[:-1] + r[1:]) / 2.0
        self.value = gr / norm
